//*****************************************************************************
//
//  File........: RTC.c
//
//  Author(s)...: ATMEL Norway
//
//  Target(s)...: ATmega169
//
//  Compiler....: AVR-GCC 3.3.1; avr-libc 1.0
//
//  Description.: Real Time Clock (RTC)
//
//  Revisions...: 1.0
//
//  YYYYMMDD - VER. - COMMENT                                       - SIGN.
//
//  20021015 - 1.0  - Created                                       - LHM
//  20031009          port to avr-gcc/avr-libc                      - M.Thomas
//  20060602		  Stripped down									- O.Schwaneberg
//  20080219          Major Modifications, added DCF-Support        - O.Schwaneberg
//
//*****************************************************************************

//  Include files
//mtA
#include <avr/io.h>
#include <avr/interrupt.h>
#include <inttypes.h>
#include <compat/deprecated.h>
#include <avr/pgmspace.h>

#include "RTC.h"

#define TRUE 1
#define FALSE 0

static uint16_t intMs;
//volatile uint8_t gSecCounter;
static uint8_t gSECOND;
static uint8_t gMINUTE;
static uint8_t gHOUR;
static volatile uint8_t  gDAY;
static volatile uint8_t  gMONTH;
static volatile uint16_t gYEAR;

static char dcfSync = 0;
static char correction = 0;

//DCFsync und MEZ Information
static char dcfMEZ = 0;
static char dcfInited = FALSE;

static uint16_t lcdOffTimer = 3600;

// Lookup table holding the length of each mont. The first element is a dummy.
// mt this could be placed in progmem too, but the arrays are accessed quite
//    often - so leaving them in RAM is better...
static char MonthLength[13] = {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

static char LeapMonth;
//static char LOCK = 0;

//Hier wird der dcf String gespeichert
static uint8_t dcfData[8] = {0,0,0,0,0,0,0,0};
static uint16_t dcfPuls = 0;
static uint16_t dcfPause = 0;
static uint8_t dcfBit = 0;
static uint8_t dcfByte = 0;


/******************************************************************************
*
*   Function name:  RTC_init
*
*   returns:        none
*
*   parameters:     none
*
*   Purpose:        Start Timer/Counter2 in asynchronous operation using a
*                   32.768kHz crystal.
*
*******************************************************************************/
void RTC_init(void)
{
    // initial time and date setting
    gSECOND  = 0;
    gMINUTE  = 0;
    gHOUR    = 0;
    gDAY     = 0;
	gMONTH   = 1;
    gYEAR    = 8;

    calcMs();
}

uint8_t getSecond(void){
	return gSECOND;
}

uint8_t getMinute(void){
	return gMINUTE;
}

uint8_t getHour(void){
	return gHOUR;
}

uint16_t getDay(void){
	return gDAY;
}

void setHour(uint8_t hour){
	gHOUR = hour;
}

void setMinute(uint8_t minute){
	gMINUTE = minute;
}

void setSecond(uint8_t second){
	gSECOND = second;
}

void setBacklightAutoOff(uint16_t seconds){
	lcdOffTimer = seconds;
	lcd_BacklightOn();
}

void calcMs(void){
	setTimeMs(((uint32_t)getHour()*3600000) + ((uint32_t)getMinute()*60000) + ((uint32_t)getSecond()*1000));
}

void parseDcfData(void){
	//VORSICHT: DCF verwendet BIG-ENDIAN!

	//Bit 0 bis 19 sind egal
	uint8_t dcfMinute = 0;
	uint8_t dcfHour = 0;
	uint8_t dcfDay = 0;
	uint8_t dcfMonth = 0;
	uint8_t dcfYear = 0;
	char    dcfCurMEZ = 0; //0=ungültig, 1= MEZ, 2=MESZ

	char parity = 0;

	//Bit 17 	„0“: MEZ, „1“: MESZ und Bit 18 	„0“: MESZ, „1“: MEZ
	if(!(dcfData[2] & 0x20) && (dcfData[2] & 0x10)){
		dcfCurMEZ = 1;
	}else if((dcfData[2] & 0x20) && !(dcfData[2] & 0x10)){
		dcfCurMEZ = 2;
	}else{
		lcd_putc('Z');
		return;
	}

	//Bit 19 (dcfData[2] & 0x08) „1“: Am Ende dieser Stunde wird eine Schaltsekunde eingefügt

	//Bit 20 ist immer 1
	if(!(dcfData[2] & 0x04)){
		//sendString("RTC: Warning: Bit 20 = 0!\n");
		lcd_putc('W');
		return;
	}

	//Bit 21-27 sind die Minute
	//Zunächst kommen Bit 21-23
	if(dcfData[2]&0x02){
		dcfMinute += 1;
		parity++;
	}
	if(dcfData[2]&0x01){
		dcfMinute += 2;
		parity++;
	}
	if(dcfData[3]&0x80){
		dcfMinute += 4;
		parity++;
	}
	if(dcfData[3]&0x40){
		dcfMinute += 8;
		parity++;
	}
	if(dcfData[3]&0x20){
		dcfMinute += 10;
		parity++;
	}
	if(dcfData[3]&0x10){
		dcfMinute += 20;
		parity++;
	}
	if(dcfData[3]&0x08){
		dcfMinute += 40;
		parity++;
	}

	//Bit 28 ist Parität
	if (dcfData[3]&0x04) {	//Prüfbit ok ?
		if (parity == 1 || parity == 3 || parity == 5 || parity == 7) {
			parity = 0;
		}
		else {
			//sendString("DCF: PARITY ERROR: Minute (1)\n");
			lcd_putc('M');
			return;
		}
	}else {							//Prüfbit ok ?
		if (parity == 0 || parity == 2 || parity == 4 || parity == 6) {
			parity = 0;
		}
		else{
			//sendString("DCF: PARITY ERROR: Minute (0)\n");
			lcd_putc('M');
			return;
		}
	}

	//Bits 29-34 kodieren die Stunde
	if(dcfData[3]&0x02){
		dcfHour += 1;
		parity++;
	}
	if(dcfData[3]&0x01){
		dcfHour += 2;
		parity++;
	}
	if(dcfData[4]&0x80){
		dcfHour += 4;
		parity++;
	}
	if(dcfData[4]&0x40){
		dcfHour += 8;
		parity++;
	}
	if(dcfData[4]&0x20){
		dcfHour += 10;
		parity++;
	}
	if(dcfData[4]&0x10){
		dcfHour += 20;
		parity++;
	}

	//Bit 35 ist Parität
	if (dcfData[4]&0x08) {	//Prüfbit ok ?
		if (parity == 1 || parity == 3 || parity == 5) {
			parity = 0;
		}
		else {
			//sendString("DCF: PARITY ERROR: Hour (1)\n");
			lcd_putc('H');
			return;
		}
	}else {							//Prüfbit ok ?
		if (parity == 0 || parity == 2 || parity == 4 || parity == 6) {
			parity = 0;
		}
		else {
			//sendString("DCF: PARITY ERROR: Hour (0)\n");
			lcd_putc('H');
			return;
		}
	}

	//Bit 36-41 Monatstag
	if(dcfData[4]&0x04){
		dcfDay += 1;
		parity++;
	}
	if(dcfData[4]&0x02){
		dcfDay += 2;
		parity++;
	}
	if(dcfData[4]&0x01){
		dcfDay += 4;
		parity++;
	}
	if(dcfData[5]&0x80){
		dcfDay += 8;
		parity++;
	}
	if(dcfData[5]&0x40){
		dcfDay += 10;
		parity++;
	}
	if(dcfData[5]&0x20){
		dcfDay += 20;
		parity++;
	}

	//Bit 42-44 Wochentag (0x10 0x08 0x04)
	if(dcfData[5]&0x10){
		parity++;
	}
	if(dcfData[5]&0x08){
		parity++;
	}
	if(dcfData[5]&0x04){
		parity++;
	}

	//Bit 45–49 	5 bit 	Monat (1, 2, 4, 8, 10)
	if(dcfData[5]&0x02){
		dcfMonth += 1;
		parity++;
	}
	if(dcfData[5]&0x01){
		dcfMonth += 2;
		parity++;
	}
	if(dcfData[6]&0x80){
		dcfMonth += 4;
		parity++;
	}
	if(dcfData[6]&0x40){
		dcfMonth += 8;
		parity++;
	}
	if(dcfData[6]&0x20){
		dcfMonth += 10;
		parity++;
	}

	//Bit 50–57 	8 bit 	Jahr (zweistellig; 1, 2, 4, 8, 10, 20, 40, 80)
	if(dcfData[6]&0x10){
		dcfYear += 1;
		parity++;
	}
	if(dcfData[6]&0x08){
		dcfYear += 2;
		parity++;
	}
	if(dcfData[6]&0x04){
		dcfYear += 4;
		parity++;
	}
	if(dcfData[6]&0x02){
		dcfYear += 8;
		parity++;
	}
	if(dcfData[6]&0x01){
		dcfYear += 10;
		parity++;
	}
	if(dcfData[7]&0x80){
		dcfYear += 20;
		parity++;
	}
	if(dcfData[7]&0x40){
		dcfYear += 40;
		parity++;
	}
	if(dcfData[7]&0x20){
		dcfYear += 80;
		parity++;
	}

	//Bit 58 	1 bit 	Parität Datum (über Bit 36 bis 57)
	if (dcfData[7]&0x10) {	//Prüfbit ok ?
		if (parity%2 == 1) {
			parity = 0;
		}
		else {
			//sendString("DCF: PARITY ERROR: Minute (1)\n");
			lcd_putc('Y');
			return;
		}
	}else {							//Prüfbit ok ?
		if (parity%2 == 0) {
			parity = 0;
		}
		else{
			//sendString("DCF: PARITY ERROR: Minute (0)\n");
			lcd_putc('Y');
			return;
		}
	}

	if(dcfInited){
		uint32_t tempTime = (((uint32_t)dcfHour)*3600000) + (((uint32_t)dcfMinute)*60000);

		if(getTimeMs() > 85500000){
			dcfSync = 30;
			return;
		}else if(getTimeMs() < 900000){
			dcfSync = 15;
			return;
		}

		//600.000ms = 10min Toleranz
		if(dcfCurMEZ == dcfMEZ){
			if((getTimeMs() > (tempTime+600000)) || (getTimeMs() < (tempTime-600000))){
				return;
			}
		}else{
			if((getTimeMs() > (tempTime+4200000)) || (getTimeMs() < (tempTime-4200000))){
				return;
			}
		}
	}

	dcfInited = TRUE;

	//Jetzt wird zugewiesen
	gHOUR = dcfHour;
	gMINUTE = dcfMinute;
	gSECOND = 0;
	gDAY = dcfDay;
	gMONTH = dcfMonth;
	gYEAR = dcfYear;
	dcfMEZ = dcfCurMEZ;

	//Die nächste Synchronisation timen
	dcfSync = 60;

	//Die Millisekunden neu berechnen
	calcMs();
}

/*
 * Die RTC muss sich dem 244,140625Hz Takt des PWM-Generators unterwerfen.
 * Das ist relativ suboptimal, aber für die Genauigkeit haben wir ja
 * das DCF-Modul.
 */
SIGNAL(SIG_OVERFLOW0)
{
	intMs += 1;

	/*
	 * Das DCF-Signal muss hier zusammen gebastelt werden
	 */
	if(dcfSync == 0){
		if(PINA & (1<<PINA1)){
		//if(ADC_read(1) > 512){
			//Wenn es eine Sekunde Pause gab fängt eine neue Phase an
			if(dcfPause > 244){
				//Haben wir ein gültiges Signal aufgenommen?
				if(dcfByte == 7 && dcfBit == 3){
					parseDcfData();
				}

				dcfBit = 0;
				dcfByte = 0;
				dcfData[0]=dcfData[1]=dcfData[2]=dcfData[3]=dcfData[4]=dcfData[5]=dcfData[6]=dcfData[7]=0;
			}
			dcfPuls++;
//			if(dcfPuls > 10){
			dcfPause = 0;
//			}
		}else{
			if(dcfPause > 44){
				if(dcfPuls > 12 && dcfPuls < 35){
					//Kurzer Puls (100ms) bedeutet logisch "0"
					dcfBit++;
					if(dcfBit > 7){
						dcfBit = 0;
						dcfByte++;
					}
				}else if(dcfPuls > 35 && dcfPuls < 73){
					//Langer Puls (200ms) bedeutet logisch "1"
					dcfBit++;
					if(dcfBit > 7){
						dcfBit = 0;
						dcfByte++;
					}
					if(dcfByte < 8){
						dcfData[dcfByte] |= (1<<(7-dcfBit));
					}
				}
				dcfPuls = 0;
			}
			dcfPause++;
		}
	}

	if(intMs == 122){
		setTimeMs(getTimeMs() + 500); //500ms hochzählen

		//Neuberechnung
		//Gültigkeit des Caches überprüfen
		cacheKeeper();

		//Intensitäten Interpolieren und speichen.
		calculateIntensities();

		//Buttonabfrage!?
//		keyHandler();
	}

	if((--correction > 0 && intMs==244)||(intMs==245)){
		intMs = 0;
		setTimeMs(getTimeMs() + 500); //500ms hochzählen

		/* Da eine Sekunde 244,140625 Schritte enthält benötigt man eine kleine Korrektur:
		 * Alle 7 Sekunden wird ein Schritt mehr abgewartet.
		 * */
		if(correction == 0){
			correction = 7;
		}

		if(lcdOffTimer > 0 && lcdOffTimer != 0xFFFF){
			lcdOffTimer--;
		}else if(lcdOffTimer == 0){
			lcd_BacklightOff();
		}

		gSECOND++;               // increment second
		if (gSECOND == 60)
		{
			gSECOND = 0;
			gMINUTE++;

			if(dcfSync > 0){
				dcfSync--;
			}

			if (gMINUTE > 59)
			{
				gMINUTE = 0;
				gHOUR++;

				if (gHOUR > 23)
				{

					gHOUR = 0;
					gDAY++;
					setTimeMs(0);

					// Check for leap year if month == February
					if (gMONTH == 2)
						if (!(gYEAR & 0x0003))              // if (gYEAR%4 == 0)
							if (gYEAR%100 == 0)
								if (gYEAR%400 == 0)
									LeapMonth = 1;
								else
									LeapMonth = 0;
							else
								LeapMonth = 1;
						else
							LeapMonth = 0;
					else
						LeapMonth = 0;

					// Now, we can check for month length
					if (gDAY > (MonthLength[gMONTH] + LeapMonth))
					{
						gDAY = 1;
						gMONTH++;

						if (gMONTH > 12)
						{
							gMONTH = 1;
							gYEAR++;
						}
					}
				}
			}
		}
//		refreshMenu();
		/*lcd_clrscr();
		lcd_puts("Prometheus-LMC\n");*/
		lcd_gotoxy(0,1);
		if(gHOUR < 10){
			lcd_putc('0');
		}
		lcd_putu(gHOUR);
		lcd_putc(':');
		if(gMINUTE < 10){
			lcd_putc('0');
		}
		lcd_putu(gMINUTE);
		lcd_putc(':');
		if(gSECOND < 10){
			lcd_putc('0');
		}
		lcd_putu(gSECOND);

		if(dcfSync == 0){
			lcd_putc(' ');
			lcd_putu(8*dcfByte + dcfBit);
			if(dcfInited)
				lcd_puts("* ");
			else
				lcd_puts("i ");
		}else{
			lcd_puts(" DCF    ");
		}

		keyHandler();
	}
}
